/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package conferencesolver.problem;

import conferencesolver.solution.ConferenceSolution;

/**
 *
 * @author Michael Brooks <mjbrooks@uw.edu>
 */
public class CachedObjective implements Objective {

    private final Objective objective;
    private final ConferenceProblem problem;
    private double[] slotValues;
    private double[][] trackValues;
    private final int sessionIncrement;
    private int cacheMisses = 0;
    private int cacheHits = 0;

    public CachedObjective(ConferenceProblem problem, Objective objective) {
        this.objective = objective;
        this.problem = problem;
        slotValues = new double[problem.getTotalPapers()];
        this.sessionIncrement = problem.getTracksInSession() * problem.getPapersInSession();

        for (int i = 0; i < slotValues.length; i++) {
            slotValues[i] = Double.NaN;
        }

        trackValues = new double[problem.getNumSessions()][problem.getTracksInSession()];

        for (int i = 0; i < problem.getNumSessions(); i++) {
            for (int j = 0; j < problem.getTracksInSession(); j++) {
                trackValues[i][j] = Double.NaN;
            }
        }
    }

    @Override
    public double getValueFromSlot(ConferenceSolution solution, int slot) {
        if (Double.isNaN(slotValues[slot])) {
            slotValues[slot] = objective.getValueFromSlot(solution, slot);
            cacheMisses++;
        } else {
            cacheHits++;
        }

        return slotValues[slot];
    }

    @Override
    public double getValueFromTrack(ConferenceSolution solution, int session1, int track1) {
        if (Double.isNaN(trackValues[session1][track1])) {
            trackValues[session1][track1] = objective.getValueFromTrack(solution, session1, track1);
            cacheMisses++;
        } else {
            cacheHits++;
        }

        return trackValues[session1][track1];
    }

    public int getCacheHits() {
        return cacheHits;
    }

    public int getCacheMisses() {
        return cacheMisses;
    }

    public void invalidate(int session) {
        int sessionStart = session * sessionIncrement;
        int sessionEnd = (session + 1) * sessionIncrement;
        for (int i = sessionStart; i < sessionEnd; i++) {
            slotValues[i] = Double.NaN;
        }
        
        for (int t = 0; t < problem.getTracksInSession(); t++) {
            trackValues[session][t] = Double.NaN;
        }
    }

    @Override
    public double scoreConference(ConferenceSolution conference) {
        return objective.scoreConference(conference);
    }

    @Override
    public double getSwapValueDelta(ConferenceSolution solution, int slot1, int slot2) {

        double scoreDelta = 0;

        scoreDelta -= getValueFromSlot(solution, slot1);
        scoreDelta -= getValueFromSlot(solution, slot2);

        //Now actually swap
        solution.swapSlots(slot1, slot2);

        //No way to speed these up
        scoreDelta += objective.getValueFromSlot(solution, slot1);
        scoreDelta += objective.getValueFromSlot(solution, slot2);

        //Now unswap
        solution.swapSlots(slot1, slot2);

        return scoreDelta;
    }

    @Override
    public double getDistance(int paperA, int paperB) {
        return objective.getDistance(paperA, paperB);
    }

    @Override
    public double getTrackSwapValueDelta(ConferenceSolution solution, int session1, int track1, int session2, int track2) {
        double scoreDelta = 0;
        
        scoreDelta -= getValueFromTrack(solution, session1, track1);
        scoreDelta -= getValueFromTrack(solution, session2, track2);

        solution.swapTracks(session1, track1, session2, track2);
        
        scoreDelta += objective.getValueFromTrack(solution, session1, track1);
        scoreDelta += objective.getValueFromTrack(solution, session2, track2);

        solution.swapTracks(session1, track1, session2, track2);
        
        return scoreDelta;
    }
}
