/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package conferencesolver.problem;

import conferencesolver.solution.ConferenceSolution;

/**
 *
 * @author Michael Brooks <mjbrooks@uw.edu>
 */
public class ConferenceObjective implements Objective {

    private final double[][] distanceMatrix;
    private double tradeoffCoefficient;

    public ConferenceObjective(double[][] distanceMatrix, double tradeoffCoefficient) {
        this.distanceMatrix = distanceMatrix;
        this.tradeoffCoefficient = tradeoffCoefficient;
    }

    @Override
    public double getDistance(int paperA, int paperB) {
        return distanceMatrix[paperA][paperB];
    }
    
    /**
     * Score the organization.
     *
     * @return the score.
     */
    @Override
    public double scoreConference(ConferenceSolution conference) {

        ConferenceProblem problem = conference.getProblem();

        // Sum of pairwise similarities per session.
        double score1 = 0.0;
        // Sum of distances for competing papers.
        double score2 = 0.0;
        for (int s = 0; s < problem.getNumSessions(); s++) {
//            Session tmpSession = conference.getSession(s);
            for (int t = 0; t < problem.getTracksInSession(); t++) {
//                Track tmpTrack = tmpSession.getTrack(t);
                for (int k = 0; k < problem.getPapersInSession(); k++) {
                    int paper1 = conference.getPaper(s, t, k);

                    for (int q = k + 1; q < problem.getPapersInSession(); q++) {
                        int paper2 = conference.getPaper(s, t, q);
                        //System.out.println("sim:"+index1+" "+index2+" "+(1-distanceMatrix[index1][index2]));
                        score1 += 1 - distanceMatrix[paper1][paper2];
                    }
                }

                //Different tracks in the same session
                for (int t2 = t + 1; t2 < problem.getTracksInSession(); t2++) {
//                    Track otherTrack = tmpSession.getTrack(t2);

                    for (int k = 0; k < problem.getPapersInSession(); k++) {
                        int paper1 = conference.getPaper(s, t, k);

                        for (int q = 0; q < problem.getPapersInSession(); q++) {
                            int paper2 = conference.getPaper(s, t2, q);
                            //System.out.println("dif:"+index1+" "+index2+" "+(distanceMatrix[index1][index2]));
                            score2 += distanceMatrix[paper1][paper2];
                        }
                    }
                }

            }
        }

        double score = score1 + tradeoffCoefficient * score2;

        return score;
    }

    @Override
    public double getSwapValueDelta(ConferenceSolution solution, int slot1, int slot2) {

        double scoreDelta = 0;

        scoreDelta -= getValueFromSlot(solution, slot1);
        scoreDelta -= getValueFromSlot(solution, slot2);

        //Now actually swap
        solution.swapSlots(slot1, slot2);

        scoreDelta += getValueFromSlot(solution, slot1);
        scoreDelta += getValueFromSlot(solution, slot2);

        //Now unswap
        solution.swapSlots(slot1, slot2);

        return scoreDelta;
    }
    
    @Override
    public double getTrackSwapValueDelta(ConferenceSolution solution, int session1, int track1, int session2, int track2) {
        double scoreDelta = 0;
        
        scoreDelta -= getValueFromTrack(solution, session1, track1);
        scoreDelta -= getValueFromTrack(solution, session2, track2);
        
        solution.swapTracks(session1, track1, session2, track2);
        
        scoreDelta += getValueFromTrack(solution, session1, track1);
        scoreDelta += getValueFromTrack(solution, session2, track2);
        
        solution.swapTracks(session1, track1, session2, track2);
        
        return scoreDelta;
    }
    
    @Override
    public double getValueFromTrack(ConferenceSolution solution, int session1, int track1) {
        double score2 = 0;
        
        ConferenceProblem problem = solution.getProblem();
        
        int track1Start = solution.getGlobalSlot(session1, track1, 0);
        
        //Different tracks in the same session
        for (int t2 = 0; t2 < problem.getTracksInSession(); t2++) {
            
            int track2Start = solution.getGlobalSlot(session1, t2, 0);
            
            if (t2 == track1) {
                continue; //skip this track
            }

            for (int q = 0; q < problem.getPapersInSession(); q++) {
                
                int otherPaperId = solution.getPaper(track2Start + q);

                if (otherPaperId == -1) {
                    continue;
                }
                
                for (int p = 0; p < problem.getPapersInSession(); p++) {
                    int paperId = solution.getPaper(track1Start + p);
                    
                    score2 += distanceMatrix[paperId][otherPaperId];
                }
            }
        }

        return tradeoffCoefficient * score2;
    }

    @Override
    public double getValueFromSlot(ConferenceSolution solution, int slot) {
        ConferenceProblem problem = solution.getProblem();

        double score1 = 0;
        double score2 = 0;

        int paperId = solution.getPaper(slot);
        int[] address = solution.getAddress(paperId);

        int sessionIdx = address[0];
        int trackIdx = address[1];
        int paperIdx = address[2];

        int trackOffset = sessionIdx * solution.getSessionIncrement() + trackIdx * solution.getTrackIncrement();
        for (int k = 0; k < problem.getPapersInSession(); k++) {
            if (k == paperIdx) {
                continue; //skip the paper we're evaluating
            }

            int otherPaperId = solution.getPaper(trackOffset + k);

            if (otherPaperId == -1) {
                continue;
            }
            if (otherPaperId == paperId) {
                throw new IllegalArgumentException("Paper already exists in that session");
            }

            //System.out.println("sim:"+index1+" "+index2+" "+(1-distanceMatrix[index1][index2]));
            score1 += 1 - distanceMatrix[paperId][otherPaperId];
        }

        //Different tracks in the same session
        trackOffset = sessionIdx * solution.getSessionIncrement();
        for (int t2 = 0; t2 < problem.getTracksInSession(); t2++, trackOffset += solution.getTrackIncrement()) {
            
            if (t2 == trackIdx) {
                continue; //skip this track
            }

            for (int q = 0; q < problem.getPapersInSession(); q++) {

                int otherPaperId = solution.getPaper(trackOffset + q);

                if (otherPaperId == -1) {
                    continue;
                }
                if (otherPaperId == paperId) {
                    throw new IllegalArgumentException("Paper already exists in another session");
                }
                //System.out.println("dif:"+index1+" "+index2+" "+(distanceMatrix[index1][index2]));
                score2 += distanceMatrix[paperId][otherPaperId];
            }
        }

        return score1 + tradeoffCoefficient * score2;
    }

}
