package conferencesolver.solution;

import conferencesolver.problem.ConferenceProblem;
import java.util.Arrays;

/**
 * Conference contains an array of sessions, each of which contains an array of
 * tracks, each of which contains an array of papers.
 *
 * @author janara
 *
 */
public class ConferenceSolution {

    // The array of tracks.
//    private Session[] sessions;
    // Stores the paper in each slot (session/track/index)
    private int[] papersInSlots;
    // Stores the slot number of each paper (ref papersInSlots)
    private int[] slotsOfPapers;
    private int sessionIncrement;
    private int trackIncrement;
    private final ConferenceProblem problem;
    
    /**
     * Constructor for Conference.
     *
     * @param numSessions is the number of sessions in a track.
     * @param parallelTracks is the number of parallel tracks.
     * @param papersInSession is the number of papers in a session.
     */
    public ConferenceSolution(ConferenceProblem problem) {
        this.problem = problem;
        this.sessionIncrement = problem.getTracksInSession() * problem.getPapersInSession();
        this.trackIncrement = problem.getPapersInSession();
        initSessions();
    }

    public int getSessionIncrement() {
        return sessionIncrement;
    }

    public int getTrackIncrement() {
        return trackIncrement;
    }
    
    
    
    
    /**
     * Initialize the tracks.
     *
     * @param numSessions is the number of sessions in a track.
     * @param parallelTracks is the number of parallel tracks.
     * @param papersInSession is the number of papers in a session.
     */
    private void initSessions() {
        papersInSlots = new int[problem.getTotalPapers()];
        slotsOfPapers = new int[problem.getTotalPapers()];

        for (int i = 0; i < papersInSlots.length; i++) {
            papersInSlots[i] = -1;
            slotsOfPapers[i] = -1;
        }

//        sessions = new Session[problem.getNumSessions()];
//        for (int s = 0; s < problem.getNumSessions(); s++) {
//            Session tempSession = new Session(problem.getTracksInSession());
//            for (int t = 0; t < problem.getTracksInSession(); t++) {
//                Track tempTrack = new Track(problem.getPapersInSession());
//                tempSession.setTrack(t, tempTrack);
//            }
//            sessions[s] = tempSession;
//        }
    }

    public ConferenceProblem getProblem() {
        return problem;
    }

    /**
     * Gets the session at the specified index.
     *
     * @param index is the index of the specified session.
     * @return the session
     */
//    public Session getSession(int index) {
//        if (index < sessions.length) {
//            return sessions[index];
//        } else {
//            throw new IndexOutOfBoundsException();
//        }
//    }
    /**
     * Sets the paper in the specified slot to the given paper id.
     *
     * @param trackIndex is the track index.
     * @param sessionIndex is the session index.
     * @param paperIndex is the paper index.
     * @param paperId is the id of the paper.
     * @throws IndexOutOfBoundsException
     */
    public void setPaper(int sessionIndex,
            int trackIndex,
            int paperIndex,
            int paperId) throws IndexOutOfBoundsException {
        int slot = paperIndex + trackIndex * trackIncrement + sessionIndex * sessionIncrement;
        setPaper(slot, paperId);

//        if (sessions.length > sessionIndex) {
//            Session curTrack = sessions[sessionIndex];
//            curTrack.setPaper(trackIndex, paperIndex, paperId);
//        } else {
//            throw new IndexOutOfBoundsException();
//        }
    }

    
    public void removePaper(int slot, int paperId) {
        if (papersInSlots.length > slot) {
            slotsOfPapers[paperId] = -1;
            papersInSlots[slot] = -1;
        } else {
            throw new IndexOutOfBoundsException();
        }
    }
    
    public void setPaper(int slot, int paperId) {
        if (papersInSlots.length > slot) {
            //Make sure to maintain consistency
            int oldSlot = slotsOfPapers[paperId];
            if (oldSlot > -1) {
                papersInSlots[oldSlot] = -1;
            }

            int displacedPaper = papersInSlots[slot];
            if (displacedPaper > -1) {
                slotsOfPapers[displacedPaper] = -1;
            }

            papersInSlots[slot] = paperId;
            slotsOfPapers[paperId] = slot;
        } else {
            throw new IndexOutOfBoundsException();
        }
    }

    public int getPaper(int slot) {
        if (papersInSlots.length > slot) {
            return papersInSlots[slot];
        } else {
            throw new IndexOutOfBoundsException();
        }
    }

    public void setPaper(int[] address, int paperId) {
        int slot = address[2] + address[1] * trackIncrement + address[0] * sessionIncrement;
        setPaper(slot, paperId);
    }

    public int getPaper(int sessionIndex,
            int trackIndex,
            int paperIndex) {
        int slot = paperIndex + trackIndex * trackIncrement + sessionIndex * sessionIncrement;
        return getPaper(slot);
//        if (sessions.length > sessionIndex) {
//            return sessions[sessionIndex].getTrack(trackIndex).getPaper(paperIndex);
//        } else {
//            throw new IndexOutOfBoundsException();
//        }
    }

    public int getPaper(int[] address) {
        int slot = address[2] + address[1] * trackIncrement + address[0] * sessionIncrement;
        return getPaper(slot);
    }
    
    
    public int getGlobalSlot(int session, int track, int paperInTrack) {
        int slot = paperInTrack + track * trackIncrement + session * sessionIncrement;
        return slot;
    }

    public int getGlobalSlot(int paperId) {
        return slotsOfPapers[paperId];
    }

    /**
     * Gets the address of a paper as [session, track, slotInTrack]
     *
     * @param paperId
     * @return
     */
    public int[] getAddress(int paperId) {
        int slot = slotsOfPapers[paperId];

        if (slot == -1) {
            return null;
        }

        return getAddressFor(slot);
    }
    
    public int getSessionFor(int slot) {
        return slot / sessionIncrement;
    }
    
    public int[] getAddressFor(int slot) {
        int[] address = new int[3];

        address[0] = slot / sessionIncrement;

        int slotsAfterSession = slot % sessionIncrement;

        address[1] = slotsAfterSession / trackIncrement;

        address[2] = slotsAfterSession % trackIncrement;
        
        return address;
    }
    
    @Override
    public String toString() {
        StringBuilder str = new StringBuilder();
        for (int i = 0; i < papersInSlots.length; i++) {
            str.append(papersInSlots[i]);
        }
        return str.toString();
    }
    
    public ConferenceSolution deepCopy() {
        ConferenceSolution copy = new ConferenceSolution(this.problem);
        copy.slotsOfPapers = Arrays.copyOf(slotsOfPapers, slotsOfPapers.length);
        copy.papersInSlots = Arrays.copyOf(papersInSlots, papersInSlots.length);
        
//        //Copy over the papers
//        for (int s = 0; s < problem.getNumSessions(); s++) {
//            for (int t = 0; t < problem.getTracksInSession(); t++) {
//                for (int k = 0; k < problem.getPapersInSession(); k++) {
//                    copy.setPaper(s, t, k, this.getPaper(s, t, k));
//                }
//            }
//        }

        return copy;
    }

    public void swapSlots(int slot1, int slot2) {
        int paper1 = getPaper(slot1);
        int paper2 = getPaper(slot2);
        
        setPaper(slot1, paper2);
        setPaper(slot2, paper1);
    }

    public void swapTracks(int session1, int track1, int session2, int track2) {
        
        int startSlot1 = getGlobalSlot(session1, track1, 0);
        int startSlot2 = getGlobalSlot(session2, track2, 0);
        
        for (int offset = 0; offset < problem.getPapersInSession(); offset++) {
            int slot1 = startSlot1 + offset;
            int slot2 = startSlot2 + offset;
            
            swapSlots(slot1, slot2);
        }
    }
}
