/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package conferencesolver.solution.improver;

import conferencesolver.problem.CachedObjective;
import conferencesolver.solution.ConferenceSolution;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Random;

/**
 *
 * @author Michael Brooks <mjbrooks@uw.edu>
 */
public class AdvancedLocalSearch extends AbstractSolutionImprover {
    private int allowedShoulderTravel = 0;
    private int tabuListLength = 0;
    private int tabuCollisions = 0;
    private int shoulderTravelled = 0;
    private double probabilityOfKick = 0;
    private Random random = null;
    private int numKicks = 0;
    private LinkedList<String> tabuList;
    private int shoulderCount = 0;

    public AdvancedLocalSearch() {
    }

    public void enableTabuList(int tabuListLength) {
        this.tabuListLength = tabuListLength;
    }

    public void enableShoulderTravel(int allowedShoulderTravel) {
        this.allowedShoulderTravel = allowedShoulderTravel;
    }

    public void enableKick(double probabilityOfKick, Random random) {
        this.probabilityOfKick = probabilityOfKick;
        this.random = random;
    }

    private final double ZERO_THRESHOLD = 0.01;

    private ArrayList<Swap> getSwaps(ConferenceSolution solution, boolean getZeros, boolean getNegative) {
        ArrayList<Swap> swaps = new ArrayList<Swap>();

        for (int slot1 = 0; slot1 < problem.getTotalPapers(); slot1++) {
            for (int slot2 = slot1 + 1; slot2 < problem.getTotalPapers(); slot2++) {

                //Need to find the delta from switching the paper in slot1 with the paper in slot2
                double newScoreDelta = objective.getSwapValueDelta(solution, slot1, slot2);

                if (getNegative
                        || (getZeros && newScoreDelta >= 0)
                        || (newScoreDelta >= ZERO_THRESHOLD)) {
                    swaps.add(new Swap(slot1, slot2, newScoreDelta));
                }
            }
        }
        return swaps;
    }

    public int getShoulderTravelled() {
        return shoulderTravelled;
    }

    public int getTabuCollisions() {
        return tabuCollisions;
    }

    public int getAllowedShoulderTravel() {
        return allowedShoulderTravel;
    }

    public int getTabuListSize() {
        return tabuListLength;
    }

    public double getProbabilityOfKick() {
        return probabilityOfKick;
    }

    public int getNumKicks() {
        return numKicks;
    }

    private Swap chooseBestSwap(ArrayList<Swap> swaps, double minAdmissableImprovement) {

        Swap selected = null;

        //Go down the list to choose the best one
        for (int i = swaps.size() - 1; i >= 0; i--) {

            Swap possibleSwap = swaps.get(i);

            //Make sure it's not in the tabu list
            if (this.tabuList.size() > 0) {

                currentSolution.swapSlots(possibleSwap.slot1, possibleSwap.slot2);
                String solutionId = currentSolution.toString();
                //Unswap
                currentSolution.swapSlots(possibleSwap.slot1, possibleSwap.slot2);

                if (this.tabuList.contains(solutionId)) {
                    tabuCollisions++;
                    continue;
                }
            }

            if (possibleSwap.deltaScore < minAdmissableImprovement) {
                break;
            }

            //We've got a usable swap
            selected = possibleSwap;
            break;
        }
        return selected;
    }

    @Override
    public SolutionImprover copy() {
        AdvancedLocalSearch copy = new AdvancedLocalSearch();
        copy.enableKick(probabilityOfKick, random);
        copy.enableShoulderTravel(allowedShoulderTravel);
        copy.enableTabuList(tabuListLength);
        return copy;
    }

    @Override
    public void internalInit() {
        tabuList = new LinkedList<String>();
        shoulderCount = 0;
        tabuCollisions = 0;
        shoulderTravelled = 0;
        this.objective = new CachedObjective(problem, objective);
    }

    
    
    @Override
    public boolean internalStep() {

        boolean getNegative = (probabilityOfKick > 0);
        boolean getZeros = (shoulderCount < allowedShoulderTravel);

        //Look at every possible switch and see which is best
        ArrayList<Swap> swaps = getSwaps(currentSolution, getZeros, getNegative);

        boolean kicking = false;
        if (probabilityOfKick > 0
                && random.nextDouble() < probabilityOfKick) {
            Collections.shuffle(swaps);
            kicking = true;
            numKicks++;
        } else {
            Collections.sort(swaps);
        }

        double minAdmissableImprovement = ZERO_THRESHOLD;
        if (this.shoulderCount < allowedShoulderTravel) {
            minAdmissableImprovement = 0;
        }
        if (kicking) {
            minAdmissableImprovement = Double.NEGATIVE_INFINITY;
        }

        Swap selectedSwap = chooseBestSwap(swaps, minAdmissableImprovement);

        if (selectedSwap != null) {

            if (!kicking && selectedSwap.deltaScore >= ZERO_THRESHOLD) {
                //It's positive, so reset the shoulder count
                this.shoulderCount = 0;
            } else if (!kicking
                    && selectedSwap.deltaScore >= 0 && this.shoulderCount < allowedShoulderTravel) {
                //If it is close to 0 add to shoulder count
                this.shoulderCount++;
                shoulderTravelled++;
            }

            currentSolution.swapSlots(selectedSwap.slot1, selectedSwap.slot2);

            double score = this.getCurrentQuality() + selectedSwap.deltaScore;

            int session1 = currentSolution.getSessionFor(selectedSwap.slot1);
            int session2 = currentSolution.getSessionFor(selectedSwap.slot2);

            CachedObjective obj = (CachedObjective) objective;
            obj.invalidate(session1);
            if (session2 != session1) {
                obj.invalidate(session2);
            }

            if (tabuListLength > 0) {
                tabuList.addLast(currentSolution.toString());
                if (tabuList.size() > tabuListLength) {
                    tabuList.removeFirst();
                }
            }
            
            this.setCurrentSolution(currentSolution);
            this.setCurrentQuality(score);
            return true;
        } else {
            return false;
        }
    }
}
